"""Package teneva, module core.transformation: transformation of TT-tensors.

This module contains the functions for orthogonalization and truncation of the
TT-tensors.

"""
import numpy as np
import scipy as sp


from .svd import matrix_svd
from .tensor import copy


def orthogonalize(Y, k=None):
    """Orthogonalize TT-tensor.

    Args:
        Y (list): TT-tensor.
        k (int): the leading mode for orthogonalization. The TT-cores 0, 1, ...,
            k-1 will be left-orthogonalized and the TT-cores k+1, k+2, ..., d-1
            will be right-orthogonalized. It will be the last mode by default.

    Returns:
        list: orthogonalized TT-tensor.

    """
    d = len(Y)
    k = d - 1 if k is None else k

    if k is None or k < 0 or k > d-1:
        raise ValueError('Invalid mode number')

    Z = copy(Y)

    for i in range(k):
        orthogonalize_left(Z, i, inplace=True)

    for i in range(d-1, k, -1):
        orthogonalize_right(Z, i, inplace=True)

    return Z


def orthogonalize_left(Y, k, inplace=False):
    """Left-orthogonalization for TT-tensor.

    Args:
        Y (list): d-dimensional TT-tensor.
        k (int): mode for orthogonalization (>= 0 and < d-1).
        inplace (bool): if flag is set, then the original TT-tensor (i.e.,
            the function argument will be transformed). Otherwise, a copy of
            the TT-tensor will be made.

    Returns:
        list: orthogonalized TT-tensor.

    """
    d = len(Y)

    if k is None or k < 0 or k >= d-1:
        raise ValueError('Invalid mode number')

    Z = Y if inplace else copy(Y)

    r1, n1, r2 = Z[k].shape
    G1 = _reshape(Z[k], (r1 * n1, r2))
    Q, R = np.linalg.qr(G1, mode='reduced')
    Z[k] = _reshape(Q, (r1, n1, Q.shape[1]))

    r2, n2, r3 = Z[k+1].shape
    G2 = _reshape(Z[k+1], (r2, n2 * r3))
    G2 = R @ G2
    Z[k+1] = _reshape(G2, (G2.shape[0], n2, r3))

    return Z


def orthogonalize_right(Y, k, inplace=False):
    """Right-orthogonalization for TT-tensor.

    Args:
        Y (list): d-dimensional TT-tensor.
        k (int): mode for orthogonalization (> 0 and <= d-1).
        inplace (bool): if flag is set, then the original TT-tensor (i.e.,
            the function argument will be transformed). Otherwise, a copy of
            the TT-tensor will be made.

    Returns:
        list: orthogonalized TT-tensor.

    """
    d = len(Y)

    if k is None or k <= 0 or k > d-1:
        raise ValueError('Invalid mode number')

    Z = Y if inplace else copy(Y)

    r2, n2, r3 = Z[k].shape
    G2 = _reshape(Z[k], (r2, n2 * r3))
    L, Q = sp.linalg.rq(G2, mode='economic', check_finite=False)
    Z[k] = _reshape(Q, (Q.shape[0], n2, r3))

    r1, n1, r2 = Z[k-1].shape
    G1 = _reshape(Z[k-1], (r1 * n1, r2))
    G1 = G1 @ L
    Z[k-1] = _reshape(G1, (r1, n1, G1.shape[1]))

    return Z


def truncate(Y, e=1.E-10, r=1.E+12, orth=True):
    """Truncate (round) TT-tensor.

    Args:
        Y (list): TT-tensor wth overestimated ranks.
        e (float): desired approximation accuracy (> 0).
        r (int, float): maximum TT-rank of the result (> 0).
        orth (bool): if the flag is set, then tensor orthogonalization will be
            performed (it is True by default).

    Returns:
        list: TT-tensor, which is rounded up to a given accuracy "e" and
        satisfying the rank constraint "r".

    """
    d = len(Y)

    if orth:
        Z = orthogonalize(Y)
        e = e / np.sqrt(d-1) * np.linalg.norm(Z[-1])
    else:
        Z = copy(Y)

    for k in range(d-1, 0, -1):
        r1, n, r2 = Z[k].shape
        G = _reshape(Z[k], (r1, n * r2))
        U, V = matrix_svd(G, e, r)
        Z[k] = _reshape(V, (-1, n, r2))
        Z[k-1] = np.einsum('ijq,ql', Z[k-1], U, optimize=True)

    return Z


def _reshape(A, n):
    return np.reshape(A, n, order='F')



"""Package teneva, module collection.tensors: various useful TT-tensors.

This module contains the collection of functions for explicit construction of
various useful TT-tensors (delta function, polynomial and others).

"""
import numpy as np


import teneva


def tensor_const(n, v=1., i_non_zero=None, I_zero=None):
    """Build tensor in the TT-format with all values equal to given number.

    Args:
        n (list, np.ndarray): shape of the tensor. It should be list or
            np.ndarray of the length "d", where "d" is a number of dimensions.
        v (float): all elements of the tensor will be equal to this value.
        i_non_zero (list, np.ndarray): optional multi-index in the form of list
            or np.ndarray of the shape d, which is not affected by zero
            multi-indices (I_zero).
        I_zero (list, np.ndarray): optional list of lists or np.ndarray of the
            shape samples x d, which relates to multi-indices, where tensor
            should be zero (this ensures that the value at index "i_non_zero"
            is not affected).

    Returns:
        list: TT-tensor.

    """
    d = len(n)
    s = abs(v) / v if abs(v) > 1.E-16 else v
    v = abs(v)**(1./d) if abs(v) > 1.E-16 else 1.
    Y = [np.ones([1, k, 1]) * v for k in n]
    Y[-1] *= s

    if I_zero is not None:
        k = 0
        for i_zero in I_zero:
            skiped = 0
            while True:
                if i_non_zero is None or i_zero[k] != i_non_zero[k]:
                    Y[k][0, i_zero[k], 0] = 0.
                    k += 1
                    break
                else:
                    k += 1
                    skiped += 1
                    if skiped > d:
                        raise ValueError('Can not set zero items')
                    if k >= d:
                        k = 0
            if k >= d:
                k = 0

    return Y


def tensor_delta(n, i, v=1.):
    """Build TT-tensor that is zero everywhere except for a given multi-index.

    Args:
        n (list, np.ndarray): shape of the tensor. It should be list or
            np.ndarray of the length "d", where "d" is a number of dimensions.
        i (list, np.ndarray): the multi-index for nonzero element. It should
            be list or np.ndarray of the length "d".
        v (float): the value of the tensor at multi-index "i".

    Returns:
        list: TT-tensor.

    """
    d = len(n)
    s = abs(v) / v if abs(v) > 1.E-16 else v
    v = abs(v)**(1./d) if abs(v) > 1.E-16 else 1.
    Y = [np.zeros([1, k, 1]) for k in n]
    for j in range(d):
        Y[j][0, i[j], 0] = v
    Y[-1] *= s
    return Y


def tensor_poly(n, shift=0., power=2, scale=1.):
    """Build TT-tensor that is polynomial like scale * (index + shift)^power.

    Args:
        n (list, np.ndarray): shape of the tensor. It should be list or
            np.ndarray of the length "d", where "d" is a number of dimensions.
        shift (float, list, np.ndarray): the shift value. It should be list or
            np.ndarray of the length "d". It may be also float value.
        power (int): the power of polynomial.
        scale (float): the scale.

    Returns:
        list: TT-tensor.

    """
    d = len(n)
    Y = []

    shift = teneva.grid_prep_opt(shift, d)

    def _get(m, j):
        return (m + shift[j])**power

    for j, k in enumerate(n):
        if j == 0:
            G = np.zeros((1, k, 2))
            for m in range(k):
                G[0, m, :] = np.array([1., _get(m, j)])

        if j > 0 and j < d-1:
            G = np.zeros((2, k, 2))
            for m in range(k):
                G[:, m, :] = np.array([
                    [1., _get(m, j)],
                    [0., 1.],
                ])

        if j == d - 1:
            G = np.zeros((2, k, 1))
            for m in range(k):
                G[:, m, 0] = np.array([_get(m, j) * scale, scale])

        Y.append(G)

    return Y



"""Package teneva, module collection.vectors: various useful QTT-vectors.

This module contains the collection of functions for explicit construction of
various useful QTT-vectors (delta function and others).

"""
import numpy as np


def vector_delta(q, i, v=1.):
    """Build QTT-vector that is zero everywhere except for a given index.

    Construct a QTT-vector of the length "2^q" with only one nonzero element in
    position "i", that is equal to a given value "v".

    Args:
        q (int): quantization level. The resulting vector will have the total
            length 2^q (> 0).
        i (int): the index for nonzero element (< 2^q). Note that "negative
            index notation" is supported.
        v (float): the value of the vector at index "i".

    Returns:
        list: TT-tensor representing the QTT-vector.

    """
    i = _index_prepare(q, i)
    ind = _index_expand(q, i)
    Y = []
    for k in range(q):
        G = np.zeros((1, 2, 1))
        G[0, ind[k], 0] = 1.
        Y.append(G)
    Y[-1][0, ind[-1], 0] = v
    return Y


def _index_expand(q, i):
    if i < 0:
        if i == -1:
            ind = [1] * q
        else:
            raise ValueError('Only "-1" is supported for negative indices.')
    else:
        ind = []
        for _ in range(q):
            ind.append(i % 2)
            i = int(i / 2)
        if i > 0:
            raise ValueError('Index is out of range.')

    return ind


def _index_prepare(q, i):
    n = 1 << q
    if i >= n or i < -n:
        raise ValueError('Incorrect index.')
    return i if i >= 0 else n + i



"""Package teneva, module core.utils: various (inner) helper functions.

This module contains some helper functions which used in other "core" modules.

"""
import numpy as np


from .maxvol import maxvol
from .maxvol import maxvol_rect


def _maxvol(A, tau=1.1, dr_min=0, dr_max=0, tau0=1.05, k0=100):
    n, r = A.shape
    dr_max = min(dr_max, n - r)
    dr_min = min(dr_min, dr_max)

    if n <= r:
        I = np.arange(n, dtype=int)
        B = np.eye(n, dtype=float)
    elif dr_max == 0:
        I, B = maxvol(A, tau0, k0)
    else:
        I, B = maxvol_rect(A, tau, dr_min, dr_max, tau0, k0)

    return I, B


def _ones(k, m=1):
    return np.ones((k, m), dtype=int)


def _reshape(A, n):
    return np.reshape(A, n, order='F')
